/*
  Test program for fl_read_image().

  The program creates a small image at the top left window position,
  reads this image with fl_read_image() with different widths and heights,
  and displays the result with a zoom factor of 4x at 4 different positions.

  The window can be resized, thus reading the image again, to see how it
  behaves (currently *very* slow on Windows with fl_read_image()).

  There is also a timer callback that resets and redraws the image at the
  lower right position every 5 seconds. This shows a potential bug (?),
  if make_current() is not used during the draw() method. The image is
  not drawn, or the space is white in my tests.

  This test file includes Roman Kantor's proposed version of
  fl_read_image_win32(). This doesn't work with normal windows, but may
  work with Fl_Offscreen's, due to the fact that fl_window is a
  *different* object type in FLTK's current implementation:
    (a) a Window (HWND) when reading directly from a window
    (b) a Bitmap (HBITMAP) when reading from an Fl_Offscreen
  This ought to be solvable somehow.

  There are some check buttons to modify the program's behavior:

   (1)	[rfw]	"read from window ?" (default: off)
		if this is off, then the original image data are displayed,
		for comparison of display times, otherwise fl_read_image()
		is used.
   (2)	[urv]	"use Roman's version (WIN only) ?" (default: off)
		you *must* also set [rfw] ON for this to work.
		The current behavior is undefined, may also crash !!!
		USE WITH CARE ! --- until we fix this ---
   (3)	[umc]	"use make_current() ?" (default: off)
		There seems to be a bug that the image is not drawn correctly,
		if it is drawn from the timer callback, whereas it is drawn
		correctly when the whole window is drawn after a resize.
		This needs investigation...

  Compile and run this program with:

  $ fltk-config --compile read_image.cxx && ./read_image

*/
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_RGB_Image.H>
#include <FL/Fl_Check_Button.H>
#include <FL/fl_draw.H>
#include <FL/x.H>
#include <FL/Fl_Printer.H>

#include <stdio.h>

#ifdef WIN32
#  include "fl_read_image_win32_test.cxx"
#endif

/*
  Change these definitions, if you want different image sizes.
  Note that IW/IC should be an integer (divisible w/o remainder).
*/
#define IW	(60)		// image width and height
#define IC	(10)		// # of image columns
#define XF	(IW/IC)		// column factor

#define ALPHA	(0)		// 0 or alpha channel value

/*
  Some global variables for easier testing...
*/

Fl_Window	*window = 0;
Fl_Offscreen	offscr = 0;	// offscreen image
Fl_Check_Button *rfw = 0;	// read from window ?
Fl_Check_Button *urv = 0;	// use Roman's version: fl_read_image_win32()
Fl_Check_Button *umc = 0;	// use make_current()

uchar		*rgb_data = 0;	// original RGB pixel data for image
Fl_RGB_Image	*img = 0;	// RGB image

class MyBox : public Fl_Box {
  int ix_,iy_,iw_,ih_;		// captured image's coordinates
  Fl_Image *img_;		// cached image
  Fl_Offscreen off_;		// if set, use offscreen

  void draw();
  void resize (int X, int Y, int W, int H);
public:
  MyBox(int x, int y, int w, int h,
      int ix, int iy, int iw, int ih,
      Fl_Offscreen offscreen = 0);
  void reset();
}; // class MyBox

void MyBox::draw() {		// draw method
  Fl_Box::draw();		// draw the box
  
  int d = ALPHA ? 4 : 3;	// image depth
  if (!img_) {			// image not (yet) chached
    uchar	*px;		// internal pixel array
    Fl_Image	*temp;		// internal image
    if (rfw->value()) {		// read from window
      px = (uchar *)malloc(d*iw_*ih_);

      // the following statement is needed *ONLY* if the box
      // is being drawn from the draw_timer callback, but not
      // if it is being drawn after a window resize.
      // Try with and w/o it !
      // This seems to be a bug on Windows only. (?)

      if (umc->value())		  // use make_current() ?
        window()->make_current(); // why is this needed ?
	
      if (off_) { // use offscreen image

        fl_begin_offscreen(off_);

#if defined(WIN32)
      if (urv->value())		// use Roman's version
        fl_read_image_win32(	// Roman's version
		px,		// pixel array
		ix_,iy_,	// window (x,y)
		iw_,ih_,	// original image size
		ALPHA);		// alpha value
      else
#endif
        fl_read_image(		// use FLTK's version
		px,		// pixel array
		ix_,iy_,	// window (x,y)
		iw_,ih_,	// original image size
		ALPHA);		// alpha value
      temp = new Fl_RGB_Image(px,iw_,ih_,d);
      
      fl_end_offscreen();
      
      } else { // use window for fl_read_image

#if defined(WIN32)
      if (urv->value())		// use Roman's version
        fl_read_image_win32(	// Roman's version
		px,		// pixel array
		ix_,iy_,	// window (x,y)
		iw_,ih_,	// original image size
		ALPHA);		// alpha value
      else
#endif
        fl_read_image(		// use FLTK's version
		px,		// pixel array
		ix_,iy_,	// window (x,y)
		iw_,ih_,	// original image size
		ALPHA);		// alpha value
      temp = new Fl_RGB_Image(px,iw_,ih_,d);

      } // use offscreen or window

    } else {			// don't read "from window"

      temp = new Fl_RGB_Image(rgb_data,IW,IW,3);

    }				// ("read from window")

    img_ = temp->copy(w(),h());	// resize and store the image
    delete temp;
    if (rfw->value())		// if read from window
      free (px);		// free pixel array
    }				// image not yet chached
  img_->draw(x(),y());		// draw the image
} // draw()

void MyBox::resize (int X, int Y, int W, int H) {
  if (img_) {			// delete cached image
    delete img_;
    img_ = 0;
  }
  Fl_Box::resize(X,Y,W,H);
} // resize()

MyBox::MyBox(int x, int y, int w, int h,
      int ix, int iy, int iw, int ih,
      Fl_Offscreen off)
    : Fl_Box(x,y,w,h) {
  ix_ = ix;
  iy_ = iy;
  iw_ = iw;
  ih_ = ih;
  off_ = off;
  img_ = 0;
  box(FL_NO_BOX);
  align(FL_ALIGN_CENTER | FL_ALIGN_INSIDE);
}

void MyBox::reset() {
  if (img_) delete img_;
  img_ = 0;
}

MyBox *box1 = 0;
MyBox *box2 = 0;
MyBox *box3 = 0;
MyBox *box4 = 0;

void print_cb (Fl_Widget *, void *win) { // print window

  Fl_Widget *g = (Fl_Widget *)win;
  Fl_Printer p;
  
  int w,h;
  float scale;

  // box1->reset(); box2->reset();
  // box3->reset(); box4->reset();

  if (!p.start_job(1)) {
    p.start_page();
    p.printable_rect(&w,&h);
    scale = float(w)/g->w();
    p.scale(scale,scale);
    p.print_widget(g);
    p.end_page();
    p.end_job();
  }
  box1->reset(); box2->reset();
  box3->reset(); box4->reset();
};

void make_offscreen(Fl_RGB_Image *image) {	// create the offscreen image
  
  int x1 = 10, y1 = 10;

  if (!offscr) {
    offscr = fl_create_offscreen(IW+20,IW+20);
    fl_begin_offscreen(offscr);			// start offscreen drawing
      fl_color(fl_lighter(FL_GREEN));		// background
      fl_rectf(0,0,IW+20,IW+20);
      image->draw(x1,y1);			// draw image to offscreen
      fl_color(FL_WHITE);
      fl_line_style(0,3);
      fl_line(x1,y1,x1+IW,y1+IW);
      fl_line(x1+IW,y1,x1,y1+IW);
      fl_color(FL_RED);
      fl_line_style(0,1);
      fl_line(x1,y1,x1+IW,y1+IW);
      fl_line(x1+IW,y1,x1,y1+IW);
      fl_color(FL_BLACK);			// reset
      fl_line_style(0);
    fl_end_offscreen();				// end offscreen drawing
  }
}

void draw_timer(void *) {
  static int n = 0;
  // if (!offscr) make_offscreen(img);
  printf("draw_timer(%d)\n",++n);
  Fl::repeat_timeout (5.0,draw_timer);
  box4->reset();
  box4->redraw();
}

int main(int argc, char **argv) {

  int i,j,x,y,w,h;
  int r,g,b;

  setvbuf (stdout,NULL,_IONBF,0);	// set stdout to unbuffered
  // printf ("\n---------------- running ----------------\n");

  // create the image (rgb_image = global)

  rgb_data = (uchar*)malloc(IW*IW*3);		// rgb image, IWxIW px.
  uchar *ip = rgb_data;				// first byte
  for (i=0; i<IW; i++) {			// set colors ...
    for (j=0; j<IW; j++) {
      r = (j/XF+1) * 255 / IC;
      g = (i/XF+1) * 255 / IC;
      b = (((i+j)/2)/XF+1) * 255 / (2*IC);
      *ip++ = r<255 ? r : 255;
      *ip++ = g<255 ? g : 255;
      *ip++ = b<255 ? b : 255;
    }
  }

  img = new Fl_RGB_Image (rgb_data,IW,IW,3);	// create local image

  window = new Fl_Window(40+IW*9,40+IW*9);

  Fl_Box *box = new Fl_Box(10,10,IW,IW);
  box->box(FL_NO_BOX);
  box->image (img);
  box->align(FL_ALIGN_CENTER | FL_ALIGN_INSIDE);

  Fl_Box *resize_box = new Fl_Box(10+IW,10+IW,window->w()-10-IW,window->h()-10-IW);
  resize_box->box(FL_NO_BOX);

  rfw = new Fl_Check_Button(20+IW, 5,150,20,"read from window ?");
  rfw->value(0);
  rfw->visible_focus(0);

  urv = new Fl_Check_Button(20+IW,30,150,20,"use Roman's version (WIN only) ?");
  urv->value(0);
  urv->visible_focus(0);

  umc = new Fl_Check_Button(20+IW,55,150,20,"use make_current() ?");
  umc->value(0);
  umc->visible_focus(0);
  
  Fl_Button *print = new Fl_Button(window->w()-90,10,80,50,"Print");
  print->callback(print_cb,window);

  window->resizable(resize_box);
  window->size_range(100,100);
  window->end();
  window->show(argc, argv);

  window->make_current();	// needed for ...
  make_offscreen(img);		// make offscreen image --> offscr

  window->begin();

    box1 = new MyBox(20+IW,  20+IW,  IW*4,    IW*4,    10,10,IW,  IW,  0);
    box2 = new MyBox(30+IW*5,20+IW,  (IW-1)*4,(IW-1)*4,10,10,IW-1,IW-1,offscr);

//  box3 = new MyBox(20+IW,  30+IW*5,(IW-2)*4,(IW-2)*4,10,10,IW-2,IW-2);
    box3 = new MyBox(20+IW,  30+IW*5,(IW-2)*4,(IW-2)*4,-5,-5,IW-2,IW-2);

//  box4 = new MyBox(30+IW*5,30+IW*5,(IW-3)*4,(IW-3)*4,10,10,IW-3,IW-3,offscr);
    box4 = new MyBox(30+IW*5,30+IW*5,(IW-3)*4,(IW-3)*4,-5,-5,IW-3,IW-3,offscr);

  window->end();

  Fl::add_timeout(5.0,draw_timer);

  return Fl::run();
}
